/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
 
package de.bsvrz.vew.engvew.verwaltung;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.engvew.engstelle.Situation;

/**
 * Klasse zur Verwaltung von Objekten vom Typ Situation von 
 * VerkehrsModellNetzen.<br>
 * 
 * Diese abstracte Klasse stellt Funktionalität zur Verfügung, um aus einem Objekt
 * vom Typ typ.netz rekursiv alle enthaltenen Objekte vom Typ typ.verkehrsModellNetz
 * abzuarbeiten und aus den Enthaltenen Mengen Elemente (Engstellen, Staus etc.)
 * zu verwalten.
 * Die Verwaltung erstreckt sich auf die Überwachung der Mengen und der Löschung
 * (Invalidierung) ihrer Elemente.
 *
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:14:30 $ / ($Author: Sans $)
 */
public abstract class VerwaltungSituationen implements MutableSetChangeListener,
  InvalidationListener
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();


  /**
   * Die Menge der verwalteten Situationen
   */
  private Map<String, Situation> _verwalteteObjekte = new HashMap<String, Situation>();

  /**
   * Read-only Zugriff auf die verwalteten Objkete
   *
   * @return Sammlung (nicht modifizierbar) der verwalteten Objekte
   */
  public Collection<Situation> getVerwalteteObjekte()
  {
    return Collections.unmodifiableCollection(_verwalteteObjekte.values());
  }

  /**
   * Erzeuge eine neue Instanz vom verwalteten Objekttyp (Supertyp <code>Situation</code>
   *
   * @param objekt das Stellvertreter-Systemobjekt vom Typ typ.situation
   * @return das erzeugete Objekt
   */
  protected abstract Situation erzeugeSituation(SystemObject objekt);

  /**
   * Die Methode <code>initialisiereVerkehrsModellNetz</code> wird von {@link #initialisiereNetz}
   * für jedes Netz (typ.netz) aufgerufen, das auch vom Typ typ.verkehrsModellNetz ist.
   *
   * @param vmn zu initialisierendes Objekt vom Typ typ.verkehrsModellNetz
   */
  protected abstract void initialisiereVerkehrsModellNetz(ConfigurationObject vmn);

  /**
   * Die Methode <code>bereinigeSituation</code> wird aufgerufen, wenn ein verwaltetes Objekt
   * aus der Menge der verwalteteten Objekte entfernt wird
   *
   * @param situation das zu entfernende Objekt
   */
  protected void bereinigeSituation(Situation situation)
  {
    // tue nichts in der Stabdardimplementierung
  }

  /**
   * Initialisierte aus dem übergebene Netz - rekursiv alle enthaltenen Objekte vom Typ
   * typ.verkehrsModellNetz
   *
   * @param netz das zu initialisierende Objekt vom Typ typ.netz
   * @return true, wenn mindestens ein Objekt vom Typ typ.verkehrsModellNetz initialisiert
   * wurde, false sonst
   */
  protected boolean initialisiereNetz(ConfigurationObject netz)
  {
    boolean result = false;

    if (!netz.isOfType("typ.netz"))
    {
      throw new IllegalArgumentException("Objekt " + netz + " ist nicht vom Typ typ.netz");
    }

    if (netz.isOfType("typ.verkehrsModellNetz"))
    {
      initialisiereVerkehrsModellNetz(netz);
      result = true;
    }

    List<SystemObject> nbt = netz.getNonMutableSet("NetzBestandTeile").getElements();
    if (nbt == null)
    {
      throw new RuntimeException("Menge \"NetzBestandTeile\" nicht vorhanden in " + netz);
    }

    for (SystemObject obj: nbt)
    {
      if (obj.isOfType("typ.netz"))
      {
        result |= initialisiereNetz((ConfigurationObject)obj);
      }
    }
    return result;
  }

  /**
   * Initialisiert die Menge der verwalteten Objekte mit den Elementen einer Menge, die zu einem
   * VerkehrsModellNetz gehört
   *
   * @param vmn Objekt vom Typ typ.verkehrsModellNetz, das die Menge enthält
   * @param mengeBezeichner Bezeichner der Menge
   */
  protected void initialisiereMenge(ConfigurationObject vmn, String mengeBezeichner)
  {
    if (!vmn.isOfType("typ.verkehrsModellNetz"))
    {
      throw new IllegalArgumentException("Objekt " + vmn + " ist nicht vom Typ typ.verkehrsModellNetz");
    }

    MutableSet menge = vmn.getMutableSet(mengeBezeichner);
    if (menge != null)
    {
      int anzahl = 0;

      for (SystemObject obj: menge.getElements())
      {
        _debug.finer("Lese Objekt: " + obj + "\t" + obj.getType() + "\t" + obj.isValid());
        if (obj.isValid())
        {
          if (initialisiereSituation(obj))
          {
            anzahl++;
          }
        }
      }

      _debug.config("Verwalte " + anzahl + " Element(e) aus der dynamischen Menge \"" +
          mengeBezeichner + "\" von " + vmn);

      menge.addChangeListener(this);
    }
    else
    {
      throw new RuntimeException("Dynamische Menge \"" +
          mengeBezeichner + "\" nicht vorhanden in " + vmn);
    }
  }

  /**
   * Initialisiert die Situationsverwaltung, wenn eine weitere Situation
   * hinzugekommen ist
   *
   * @param obj das Object vom Typ <code>Situation</code>, dass hinzugekommen ist
   * @return true, wenn das Objekt in die Verwaltung aufgenommen wurde,
   * false sonst
   */
  private boolean initialisiereSituation(SystemObject obj)
  {
    if (_verwalteteObjekte.get(obj.getPid()) == null)
    {
      Situation situation = erzeugeSituation(obj);

      if (situation != null)
      {
        synchronized(this)
        {
          _verwalteteObjekte.put(obj.getPid(), situation);
          if (obj instanceof DynamicObject)
          {
            ((DynamicObject)obj).addListenerForInvalidation(this);
          }
        }
      }

      return (situation != null);
    }
    else
    {
      _debug.warning("Objekt mit PID " + obj.getPid() +
          " wird bereits verwaltet");
      return false;
    }
  }

  /**
   * Bereinigt die Engstellenverwaltung, wenn eine Engstelle entfernt wird
   *
   * @param obj das Object vom Typ <code>Engstelle</code>, dass entfernt wurde
   * @return true, wenn eine Bereinigung durchgeführt wurde, false sonst
   */
  private boolean bereinigeSituation(SystemObject obj)
  {
    if (_verwalteteObjekte.get(obj.getPid()) != null)
    {
      synchronized(this)
      {
        if (obj instanceof DynamicObject)
        {
          ((DynamicObject)obj).removeListenerForInvalidation(this);
        }
        Situation situation = _verwalteteObjekte.remove(obj.getPid());
        if (situation != null)
        {
          bereinigeSituation(situation);
        }
      }

      return true;
    }
    else
    {
      return false;
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet, de.bsvrz.dav.daf.main.config.SystemObject[], de.bsvrz.dav.daf.main.config.SystemObject[])
   */
  public void update(MutableSet set, SystemObject[] addedObjects,
      SystemObject[] removedObjects)
  {
    _debug.fine("MutableSetChangeListener.update() von \"" + set.getName() +"\"");

    if (addedObjects.length > 0)
    {
      _debug.config("Dynamische Menge \"" + set.getName() + "\" um " +
          addedObjects.length + " Objekte vergrössert");

      for(SystemObject obj: addedObjects)
      {
        initialisiereSituation(obj);
      }
    }

    if (removedObjects.length > 0)
    {
      _debug.config("Dynamische Menge \"" + set.getName() + "\" um " +
          removedObjects.length + " Objekte verkleinert");

      for(SystemObject obj: removedObjects)
      {
        bereinigeSituation(obj);
      }
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.config.InvalidationListener#invalidObject(de.bsvrz.dav.daf.main.config.DynamicObject)
   */
  public void invalidObject(DynamicObject dynamicObject)
  {
    _debug.info("Objekt " + dynamicObject +
        " gelöscht/invalidiert => wird nicht weiter verwaltet");
    
    bereinigeSituation(dynamicObject);
  }
    
}
